################################################################################
#
# Andy Long (ael), February, 2023
#
# Started with the code that Yoseph provided me last week -- thanks Yoseph!
#
# The tricky part is pulling the audio from the mic, and then passing it
# (altered appropriately by the audiogram) through the speakers. 
#
################################################################################
# 
# Essential libraries:
# 
######################################################################
import pyaudio
import numpy as np

import matplotlib.pyplot as plt

from scipy.fft import fft # https://docs.scipy.org/doc/scipy/tutorial/fft.html
from scipy.fft import ifft # https://docs.scipy.org/doc/scipy/tutorial/fft.html
from scipy.interpolate import CubicHermiteSpline

################################################################################
#
# Defaults/Globals
#
# ael: in the original implementation these were the defaults:
# this is the default sample rate for my iMac, per
# p.get_default_input_device_info()
# p.get_default_output_device_info()
# sampleRate=44100
sampleRate=44100
# ael: this was a choice made in Yoseph's original code:
chunkSize=2**14
print("chunkSize =",chunkSize)
print("frequency resolution  =",sampleRate/chunkSize)
chunkHalf=int(chunkSize/2)
normalization=2**17
vMax=110

# ael: this is also a characteristic of my devices, I believe:
dataType=np.int16
# lea

################################################################################
#
# We're going to pull an audiogram from our testing procedure, as a 
# function (more than likely). At the moment I create a vector of "weights" for
# the audiogram; ultimately, these will reflect both the amplitude (power) of a
# particular frequency, as well as the frequency itself, and will be read from
# storage. 
#
######################################################################
#
# This was my initial testing result using an early version of our testing procedure:
#
power = np.array([0.004410944000000001, 0.0016430319999999998, 0.0024061040000000005,
         0.0031121359999999993, 0.004251528, 0.003944312, 0.008242408000000001,
         0.0025154560000000005, 0.0016430319999999998, 0.0028632879999999994,
         0.004096000000000001, 0.008000000000000002, 0.0058319999999999995,
         0.003796416, 0.026463591999999998, 0.011852352000000002, 0.012167,
         0.025412183999999997, 0.046655999999999996, 0.054872000000000004,
                  0.06400000000000002, 0.0, 0.0, 0.0])
hz = np.array([125.0, 250.0, 500.0, 1000.0, 1062.5, 1562.5, 2000.0, 2500.0, 2562.5,
      3000.0, 3250.0, 3500.0, 3750.0, 4000.0, 5000.0, 5500.0, 6000.0, 6500.0,
      7000.0, 7500.0, 8000.0,8500,9000,9500])

# Let's scale up that power so that it seems more decibelish:
power = 1000 * power

# Calculate the first derivatives of the data points,
dydx = np.gradient(power,hz)
# and then create a cubic Hermite spline interpolant:
interp = CubicHermiteSpline(hz, power, dydx)

def Audiogram(freq):
    return interp(freq)
def tooBig(x):
    return vMax if x > vMax else x
def tooSmall(x):
    return 0 if x < 0 else x

# 
# Here we define our compression function, which was fine-tuned using Mathematica file WDRC.nb:
# 
def sigmoidal(a, x):
    return 1 - 1 / (1 + np.exp(x / a))
def l(tf, v):
    return tf + (vMax - tf) / vMax * v
def drc(tf, v):
    return sigmoidal(0.25, (v - tf/20)) * sigmoidal(18, (tf/18 - (v - 104))) * l(tf, v) + vMax * sigmoidal(18, (-tf/18 + (v - 104)))

#
# Let's plot the Audiogram:
#
f = np.logspace(np.log10(125), np.log10(10000), num=1000)
plt.plot(hz, power, 'o', label='data')
plt.plot(f, [tooSmall(tooBig(Audiogram(x))) for x in f])
plt.xscale('log')
plt.xlabel('log(freq)')
plt.ylabel('threshold volume')
plt.title('Cubic Spline-Fit Audiogram from Testing')
plt.show()

#
# Here's an example of the WDRC function:
#
v = np.linspace(0, 120, 1000)
plt.plot(v, drc(20, v))
plt.xlabel('input volume')
plt.ylabel('compressed volume')
plt.title('For a threshold of 20dB')
plt.show()

#
# Let's create that matrix -- notice the "backward" indexing (from what I would expect, anyway):
#
freqz = np.arange(0, chunkHalf+1, 1) * sampleRate / chunkSize
volz = np.arange(0, 111, 1)
WDRC = np.zeros((len(volz), len(freqz)))
for i in range(len(volz)):
    for j in range(len(freqz)):
        WDRC[i, j] = drc(tooSmall(tooBig(Audiogram(freqz[j]))), volz[i])

#
# Now let's plot the WDRC matrix:
#
freqz, volz = np.meshgrid(freqz, volz)
fig = plt.figure()
ax = fig.add_subplot(projection='3d')
ax.plot_surface(freqz, volz, WDRC, cmap='viridis')
ax.set_xlabel('Frequency')
ax.set_ylabel('Input Volume')
ax.set_zlabel('Output Volume')
plt.title('HLP Matrix from Given Audiogram and WDRC')
plt.show()

################################################################################
#
# Open a pyaudio channel and input and output streams:
#
# Thanks: https://people.csail.mit.edu/hubert/pyaudio/docs/
#
######################################################################
p = pyaudio.PyAudio()
streamIN  = p.open(format=pyaudio.paInt16, 
                   channels=1, 
                   rate=sampleRate, 
                   input=True, 
                   frames_per_buffer=chunkSize)
streamOUT = p.open(format=pyaudio.paInt16, 
                   channels=1, 
                   rate=sampleRate, 
                   output=True)

################################################################################
#
# We create the infinite loop, that will pull audio in, and then push it out.
#
######################################################################
def real_time_fft():
    
    while True:

        # read data from streamIN, and convert it to ints:
        data = streamIN.read(chunkSize,exception_on_overflow = False)
        origData = np.frombuffer(data, dtype=dataType)
        
        # We compute the FFT (Fast Fourier Transform):
        fft_data=fft(origData)

        # ael:
        # The original code from ChatGPT (via Yoseph) included a normalization,
        # normalization16=32768.0 # = 2^(15)
        # which appears to be one power less than the integer size. 
        # 
        # I've reinstituted it, because I want to get the magnitudes down where they look decibelish:
        fft_data_magnitudes=np.abs(fft_data) / normalization
        # If one uncomments the line below, you'll see values up into the 100s....
        print(fft_data_magnitudes[7:10])
        print(np.max(fft_data_magnitudes))
        # 

        energy=np.mean(fft_data_magnitudes)
        for i in range(1,chunkHalf+1,1): # let's compress that data:

            # 
            # We're after the WDRC value of the current input volume:
            # 
            jndex=chunkHalf-i
            index=int(tooSmall(tooBig(fft_data_magnitudes[i])))
            # it's    [volume, frequency]
            # tmp = WDRC[index , jndex    ]
            tmp = WDRC[index , jndex    ]/vMax 
            # print(tmp)

            # 
            # Now we scale the fft_data by the appropriate wdrc:
            # 
            fft_data[i]= tmp*fft_data[i]
            # ael:
            # It is far out that you can index from the end with negative indices:
            fft_data[-i] = tmp*fft_data[-i]
            # lea
            
        # fft_data_magnitudes=np.abs(fft_data) / normalization
        # print(fft_data_magnitudes[1:chunkSize])
        # asdf
        # 
        # Then we mess with the FFT, weighting the powers by the audiogram, and
        # computing its inverse fourier transform (ift):
        ift=ifft(fft_data)
        # print(ift[1:4])

        # This was the trick to writing it to the speakers: it has to be
        # converted appropriately, first to int16, then to bytes:
        ift=np.array(ift.real, dtype='int16').tobytes()

        # now write the "improved" sound to streamOUT:
        streamOUT.write(ift)

################################################################################
#
# Let 'er rip:
#
######################################################################
real_time_fft()

######################################################################
# 
# 
#  This from Chat GPT (Mar 21, 2023):
# 
# The FFT (Fast Fourier Transform) is an algorithm used to compute the Discrete
# Fourier Transform (DFT) of a discrete-time signal. The output of the FFT is a
# set of complex numbers representing the amplitude and phase of the frequency
# components of the input signal.  
# 
# The frequencies in the FFT output correspond to the frequency components of the
# input signal. The frequency resolution of the FFT is determined by the length
# of the input signal and the sampling rate used to acquire the signal. The
# frequency resolution is the smallest frequency difference that can be
# distinguished by the FFT. 
# 
# For a signal with a length of N samples and a sampling rate of Fs, the
# frequency resolution of the FFT is given by Fs/N. The output of the FFT
# contains N/2+1 frequency bins, where the first bin corresponds to DC (0 Hz) and
# the last bin corresponds to Fs/2 Hz. The frequencies of the other bins are
# evenly spaced between these two values. 
# 
# For example, if a signal is sampled at 1000 Hz and the FFT length is 1024
# samples, the frequency resolution of the FFT is 1000/1024 = 0.9766 Hz. The
# output of the FFT would contain 513 frequency bins, with the first bin
# representing DC (0 Hz) and the last bin representing 500 Hz. The frequencies of
# the other bins would be spaced at intervals of 0.9766 Hz. 
# 
######################################################################
# 
# So in our case, the sample rate Fs=44100, and FFT length is 1024; hence,
# 
# 513 frequency bins, with 
# frequency resolution 
# the last bin representing 22050/1024*513=11047 Hz. 
# 
######################################################################
